# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project
"""
This module is named `utils_` instead of `utils` to avoid obscuring
`tests/utils.py`.
"""
